"""
CLI for m3gnet
"""

import argparse
import sys
import logging
import os

from pymatgen.core.structure import Structure
import tensorflow as tf
from m3gnet.models import Relaxer

logging.captureWarnings(True)
tf.get_logger().setLevel(logging.ERROR)


def relax_structure(args):
    """
    Handle view commands.

    :param args: Args from command.
    """

    for fn in args.infile:
        s = Structure.from_file(fn)

        if args.verbose:
            print("Starting structure")
            print(s)
            print("Relaxing...")
        relaxer = Relaxer()
        relax_results = relaxer.relax(s)
        final_structure = relax_results["final_structure"]

        if args.suffix:
            basename, ext = os.path.splitext(fn)
            outfn = f"{basename}{args.suffix}{ext}"
            final_structure.to(filename=outfn)
            print(f"Structure written to {outfn}!")
        elif args.outfile is not None:
            final_structure.to(filename=args.outfile)
            print(f"Structure written to {args.outfile}!")
        else:
            print("Final structure")
            print(final_structure)

    return 0


def main():
    """
    Handle main.
    """
    parser = argparse.ArgumentParser(
        description="""
    This script works based on several sub-commands with their own options. To see the options for the
    sub-commands, type "m3g sub-command -h".""",
        epilog="""Author: M3Gnet""",
    )

    subparsers = parser.add_subparsers()

    p_relax = subparsers.add_parser("relax", help="Relax crystal structures.")

    p_relax.add_argument(
        "-i",
        "--infile",
        dest="infile",
        nargs="+",
        required=True,
        help="Input file containing structure. Common structures support by pmg.Structure.from_file method.",
    )

    p_relax.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        default=False,
        action="store_true",
        help="Verbose output.",
    )

    groups = p_relax.add_mutually_exclusive_group(required=False)
    groups.add_argument(
        "-s",
        "--suffix",
        dest="suffix",
        help="Suffix to be added to input file names for relaxed structures. E.g., _relax.",
    )

    groups.add_argument(
        "-o",
        "--outfile",
        dest="outfile",
        help="Output filename.",
    )

    p_relax.set_defaults(func=relax_structure)

    args = parser.parse_args()

    try:
        getattr(args, "func")
    except AttributeError:
        parser.print_help()
        sys.exit(-1)
    return args.func(args)


if __name__ == "__main__":
    main()



import os
import unittest

import numpy as np
from ase import Atoms
from monty.tempfile import ScratchDir
from pymatgen.core.structure import Lattice, Molecule, Structure

from m3gnet.models import M3GNet, MolecularDynamics, Potential, Relaxer, M3GNetCalculator


class TestModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.model = M3GNet.load()
        cls.potential = Potential(model=cls.model)
        cls.mol = Molecule(["C", "O"], [[0, 0, 0], [1.5, 0, 0]])
        cls.structure = Structure(Lattice.cubic(3.30), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        cls.atoms = Atoms(["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]], cell=np.eye(3) * 3.30, pbc=True)

    def test_m3gnet(self):

        g = self.model.graph_converter(self.mol)

        val = self.model.predict_structure(self.mol).numpy().ravel()
        val_graph = self.model.predict_graph(g).numpy().ravel()

        self.assertTrue(val.size == 1)
        self.assertAlmostEqual(val, val_graph)

        vals = self.model.predict_structures([self.mol, self.mol]).numpy().ravel()
        vals_graph = self.model.predict_graphs([g, g]).numpy().ravel()

        self.assertTrue(np.allclose(vals, [val, val]))
        self.assertTrue(np.allclose(vals_graph, [val, val]))

    def test_potential(self):
        e, f, s = self.potential.get_efs(self.structure)
        self.assertAlmostEqual(e.numpy().item(), -21.3307, 3)
        self.assertTrue(np.allclose(f.numpy().ravel(), np.zeros(shape=(2, 3)).ravel(), atol=1e-3))
        self.assertTrue(
            np.allclose(
                np.diag(s[0].numpy()).ravel(),
                np.array([[28.519585, 28.519585, 28.519585]]),
                atol=1e-2,
            )
        )

    def test_relaxer(self):
        relaxer = Relaxer()  # this loads the default model

        relax_results = relaxer.relax(self.structure)

        final_structure = relax_results["final_structure"]
        final_energy = relax_results["trajectory"].energies[-1] / 2
        self.assertAlmostEqual(final_structure.lattice.abc[0], 3.169, 2)
        self.assertAlmostEqual(final_energy.item(), -10.859, 3)

    def test_md(self):
        with ScratchDir("."):
            md = MolecularDynamics(atoms=self.structure, temperature=300, logfile="mo.log")

            md.run(10)
            self.assertTrue(os.path.isfile("mo.log"))

    def test_calculator(self):
        atoms = self.atoms.copy()
        atoms.calc = M3GNetCalculator(potential=self.potential)

        energy = atoms.get_potential_energy()
        forces = atoms.get_forces()

        self.assertAlmostEqual(energy, 325.37271, 3)
        self.assertTrue(
            np.allclose(
                forces[0],
                [-673.525, -673.525, -673.525],
                atol=1e-2,
            )
        )

        self.assertEqual(np.shape(energy), ())
        self.assertEqual(np.shape(forces), (2, 3))


if __name__ == "__main__":
    unittest.main()



"""
atomic reference. Used for predicting extensive properties.
"""

import logging
from typing import List, Optional

import numpy as np
import tensorflow as tf
from ase import Atoms
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Molecule, Structure

from m3gnet.config import DataType
from m3gnet.graph import Index
from m3gnet.utils import get_segment_indices_from_n, register

logger = logging.getLogger(__name__)


@register
class BaseAtomRef(tf.keras.layers.Layer):
    """
    Base AtomRef that predicts 0 correction
    """

    def call(self, graph: List, **kwargs):
        """

        Args:
            graph (list): list repr of a graph
            **kwargs:

        Returns: 0
        """
        return 0.0


@register
class AtomRef(BaseAtomRef):
    """
    Atom reference values. For example, if the average H energy is -20.0, and
    the average O energy is -10.0, the AtomRef predicts -20 * 2 + (-10.) =
    -50.0 for the atom reference energy for H2O
    """

    def __init__(
        self,
        property_per_element: Optional[np.ndarray] = None,
        max_z: int = 94,
        **kwargs,
    ):
        """
        Args:
            property_per_element (np.ndarray): element reference value
            max_z (int): maximum atomic number
            **kwargs:
        """
        super().__init__(**kwargs)
        if property_per_element is None:
            self.property_per_element = np.zeros(shape=(max_z + 1,))
        else:
            self.property_per_element = np.array(property_per_element).ravel()
        self.max_z = max_z

    def _get_feature_matrix(self, structs_or_graphs):
        n = len(structs_or_graphs)
        features = np.zeros(shape=(n, self.max_z + 1))
        for i, s in enumerate(structs_or_graphs):
            if isinstance(s, (Structure, Molecule)):
                atomic_numbers = [i.specie.Z for i in s.sites]
            elif isinstance(s, (list, np.ndarray)):
                atomic_numbers = s
            elif isinstance(s, Atoms):
                atomic_numbers = s.get_atomic_numbers()
            else:
                atomic_numbers = s.atoms[:, 0]
            features[i] = np.bincount(atomic_numbers, minlength=self.max_z + 1)
        return features

    def fit(self, structs_or_graphs, properties):
        """
        Fit the elemental reference values for the properties
        Args:
            structs_or_graphs (list): list of graphs or structures
            properties (np.ndarray): array of extensive properties
        Returns:
        """
        features = self._get_feature_matrix(structs_or_graphs)
        self.property_per_element = np.linalg.pinv(features.T.dot(features)).dot(features.T.dot(properties))
        string_prop = ""
        for i, j in enumerate(self.property_per_element):
            if abs(j) > 1e-5:
                string_prop += f"{str(Element.from_Z(i))}: {j:.5f}"
        logger.info("Property offset values: " + string_prop)
        return True

    def transform(self, structs_or_graphs, properties):
        """
        Correct the extensive properties by subtracting the atom reference
        values
        Args:
            structs_or_graphs (list): list of graphs or structures
            properties (np.ndarray): array of extensive properties
        Returns: corrected property values

        """
        properties = np.array(properties)
        atom_properties = self.predict_properties(structs_or_graphs)
        return properties - np.reshape(atom_properties, properties.shape)

    def inverse_transform(self, structs_or_graphs, properties):
        """
        Take the transformed values and get the original values
        Args:
            structs_or_graphs (list): list of graphs or structures
            properties (np.ndarray): array of extensive properties
        Returns: corrected property values

        """
        properties = np.array(properties)
        property_atoms = self.predict_properties(structs_or_graphs)
        final_properties = properties + np.reshape(property_atoms, properties.shape)
        return final_properties

    def predict_properties(self, structs_or_graphs):
        """
        Args:
            structs_or_graphs (list): calculate the atom summed property
                values
        Returns:

        """
        if not isinstance(structs_or_graphs, list):
            structs_or_graphs = [structs_or_graphs]
        features = self._get_feature_matrix(structs_or_graphs)
        return features.dot(self.property_per_element)

    def call(self, graph: List, **kwargs):
        """
        Args:
            graph (list): a list repr of a graph
            **kwargs:
        Returns:
        """
        atomic_numbers = graph[Index.ATOMS][:, 0]
        atom_energies = tf.gather(tf.cast(self.property_per_element, DataType.tf_float), atomic_numbers)
        res = tf.math.segment_sum(atom_energies, get_segment_indices_from_n(graph[Index.N_ATOMS]))
        return tf.reshape(res, (-1, 1))

    def set_property_per_element(self, property_per_element):
        """
        Set the property per atom value
        Args:
            property_per_element (np.ndarray): array of elemental properties,
                the i-th row is the elemental value for atomic number i.
        Returns:
        """
        self.property_per_element = property_per_element

    def get_config(self):
        """
        Get dict config for serialization
        Returns (dict):
        """
        config = super().get_config()
        config.update(**{"property_per_element": self.property_per_element, "max_z": self.max_z})
        return config



import json
import os
import unittest

import numpy as np
import tensorflow as tf
from m3gnet.models import M3GNet, Potential
from m3gnet.trainers import PotentialTrainer, Trainer
from monty.tempfile import ScratchDir
from pymatgen.core import Structure

DIR = os.path.dirname(os.path.abspath(__file__))


class TestTrainer(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        with open(os.path.join(DIR, "Li-O.json")) as f:
            data = json.load(f)

        cls.structures = []
        cls.energies = []
        cls.bandgaps = []
        for j in data.values():
            cls.structures.append(Structure.from_dict(j["structure"]))
            cls.energies.append(j["energy"])
            cls.bandgaps.append(j["band_gap"])

    def test_train_with_state_and_batches(self):
        """
        This error only occurs when include_states=True and the batch size
        is less than the number of data points you are training on (which
        is typically the case)
        """
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=True, include_states=True)
        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-2))

        # 50 > 32
        trainer.train(self.structures[:50], self.bandgaps[:50], batch_size=32, epochs=2, train_metrics=["mae"])
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_bandgap(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=True)
        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-2))

        trainer.train(self.structures[:30], self.bandgaps[:30], epochs=2, train_metrics=["mae"])
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_energy(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=False)

        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-3))

        with ScratchDir("."):
            trainer.train(
                self.structures[:30],
                self.energies[:30],
                epochs=2,
                train_metrics=["mae"],
            )
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_potential(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=False)
        potential = Potential(model=m3gnet)

        trainer = PotentialTrainer(potential=potential, optimizer=tf.keras.optimizers.Adam(1e-3))

        n_atoms = [len(i) for i in self.structures]
        fake_forces = [np.random.normal(size=(i, 3)) for i in n_atoms]
        fake_stress = [np.random.normal(size=(3, 3)) for i in n_atoms]
        with ScratchDir("."):
            trainer.train(
                self.structures[:30],
                self.energies[:30],
                fake_forces[:30],
                fake_stress[:30],
                validation_graphs_or_structures=self.structures[30:40],
                val_energies=self.energies[30:40],
                val_forces=fake_forces[30:40],
                val_stresses=fake_stress[30:40],
                epochs=2,
                fit_per_element_offset=True,
                save_checkpoint=False,
            )
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))

    def test_train_energy_offset(self):
        m3gnet = M3GNet(n_blocks=1, units=5, is_intensive=False)

        trainer = Trainer(model=m3gnet, optimizer=tf.keras.optimizers.Adam(1e-3))

        trainer.train(
            self.structures[:30],
            self.energies[:30],
            epochs=2,
            train_metrics=["mae"],
            fit_per_element_offset=True,
        )
        self.assertTrue(m3gnet.predict_structures(self.structures[:2]).numpy().shape == (2, 1))


if __name__ == "__main__":
    unittest.main()
